# -*- coding: utf-8 -*-

"""
Scenario.

Wrapper class for species and gene tree. Compute statistics, BMG/RBMG.
"""

import networkx as nx

from asymmetree.simulator import TreeSimulator as ts
from asymmetree.simulator.DistanceMatrix import distance_matrix
from asymmetree.best_matches import TrueBMG


__author__ = "David Schaller"
__copyright__ = "Copyright (C) 2019, David Schaller"


class Scenario:
    
    def __init__(self, S, TGT, DLH_rates, OGT=None):
        self.S = S
        self.TGT = TGT
        self.DLH_rates = DLH_rates
        
        self.OGT = OGT if OGT else ts.observable_tree(TGT)
        
        self._genes()
        self._count_events()
        self._sort_species_to_subtrees()
        self._sort_genes_to_species()
        
        self.TOG = TrueBMG.true_orthology_graph(self.OGT)
        self.BMG, self.RBMG = TrueBMG.best_match_graphs(self.OGT)
        
        self.BMG_subtrees, self.RBMG_subtrees = self.reduce_to_subtrees(self.BMG, self.RBMG)
        
        
    def _genes(self):
        
        self.OGT.supply_leaves()
        color_dict = {}
        for gene in self.OGT.root.leaves:
            if gene.color not in color_dict:
                color_dict[gene.color] = []
            color_dict[gene.color].append(gene)
        self.genes = []                                 # color-sorted gene list
        for color, gene_list in color_dict.items():
            for gene in gene_list:
                self.genes.append(gene)
                                                        # maps gene to index in matrix
        self.gene_index = {gene: i for i, gene in enumerate(self.genes)}
    
    
    def _count_events(self):
        
        self.event_counts = [self.S.number_of_species, 0, 0, 0, 0]    # count S / D / L / H /
                                                                      # ancient duplications                
        for v in self.TGT.preorder():
            if v.label == "D":
                self.event_counts[1] += 1
                if self.S.root.ID == v.color[0]:
                    self.event_counts[4] += 1
            elif v.label == "*":
                self.event_counts[2] += 1
            elif v.label == "H":
                self.event_counts[3] += 1
    
    
    def _sort_species_to_subtrees(self):
        
        # sort the species into the subtrees of the first speciation event of S
        self.subtree_list, self.subtree_index = [], {}
        for u in self.S.preorder():                          # exclude root of planted tree
            if len(u.children) > 1:
                for i in range(len(u.children)):
                    self.subtree_list.append([gene.ID for gene in self.S.preorder(node=u.children[i]) 
                                         if len(gene.children) == 0])
                    self.subtree_index.update({item: i for item in self.subtree_list[i]})
                break
        
        # max. number of outgroup genes (in T) for each subtree of S
        self.outgroup_dict = {i: [] for i in range(len(self.subtree_list))}
        for gene in self.genes:
            for i in self.outgroup_dict.keys():
                if self.subtree_index[gene.color] != i:
                    self.outgroup_dict[i].append(gene)
    
    
    def _sort_genes_to_species(self):
        
        self.color_dict = {item: [] for item in self.subtree_index.keys()}
        for gene in self.genes:
            self.color_dict[gene.color].append(gene)
    
    
    def get_data(self):
        
        return (self.S, self.genes, self.gene_index,
                self.subtree_list, self.subtree_index,
                self.outgroup_dict, self.color_dict)
    
    
    def rates_and_counts(self):
        """Return event rates and counts."""
        
        return list(self.DLH_rates) + self.event_counts
    
    
    def reduce_to_subtrees(self, full_BMG, full_RBMG):
        """Return a subgraph of the true RBMG with edges {u,v} for which the
        corresponding species (colors) are in the same subtree of root(S)."""
        BMG_subtrees = nx.DiGraph()
        RBMG_subtrees = nx.Graph()
        
        for G_subtrees, true_G in [(BMG_subtrees, full_BMG),
                                   (RBMG_subtrees, full_RBMG)]:
            G_subtrees.add_nodes_from(true_G.nodes(data=True))
            for u, v in true_G.edges:
                u_col = true_G.nodes[u]['color']
                v_col = true_G.nodes[v]['color']
                
                if self.subtree_index[u_col] == self.subtree_index[v_col]:
                    G_subtrees.add_edge(u, v)
        
        return BMG_subtrees, RBMG_subtrees
    
    
    def get_distance_matrix(self):
        _, _, D = distance_matrix(self.OGT, leaves=self.genes,
                                  leaf_index=self.gene_index)
        return D
    
    
    def possible_edges_BMG(self):
        """Return the number of possible edges in the BMG i.e. for gene pairs for
        which an outgroup is available in S."""
        
        counts = [0 for i in range(len(self.subtree_list))]
        
        for g in self.genes:
            counts[self.subtree_index[g.color]] += 1
        
        possible_edges = 0
        for count in counts:
            possible_edges += (count * (count-1))
        
        return possible_edges


# -*- coding: utf-8 -*-

"""
Orthology graph, cBMG, cRBMG.

This module provides classes concerning colored graphs and colored
(phylogentic) trees, including a cBMG(PhyloTree)-function.

Methods in this module:
    - true_orthology_graph(tree)
    - best_match_graphs(tree)
    - build_cBMG_alternative(tree)
    - RBMG_from_BMG(cBMG)
"""

import itertools

import numpy as np
import networkx as nx


__author__ = "David Schaller"
__copyright__ = "Copyright (C) 2019, David Schaller"


# --------------------------------------------------------------------------
#                 True Orthology Graph (TOG), cBMG & cRBMG
#
#                         (given a known tree)
# --------------------------------------------------------------------------

def true_orthology_graph(tree):
    """Constructs the true orthology graph from an event-labeled tree."""
    
    tree.supply_leaves()                                # assign list of leaves to each node
    G = nx.Graph()
    
    for v in tree.root.leaves:
        G.add_node(v.ID, label=v.label, color=v.color)
    
    for node in tree.preorder():
        if node.label == "S":
            for child1, child2 in itertools.combinations(node.children, 2):
                for u in child1.leaves:
                    for v in child2.leaves:
                        G.add_edge(u.ID, v.ID)
    return G


def best_match_graphs(tree):
    """Create an n-colored BMG and RBMG from a given tree.
    
    Own algorithm."""
    
    tree.supply_leaves()                                # assign list of leaves to each node
    BMG, RBMG = nx.DiGraph(), nx.Graph()
    colors = set()
    
    for v in tree.root.leaves:
        colors.add(v.color)
        BMG.add_node(v.ID, label=v.label, color=v.color)
        RBMG.add_node(v.ID, label=v.label, color=v.color)
    
    # ---- build BMG ----
    for u in tree.root.leaves:
        remaining = colors - set([u.color])             # colors to which no best match has yet been found
        parent = u.parent                               # start with direct parent of each node
        while remaining and parent:
            colors_here = set()
            for v in parent.leaves:
                if v.color in remaining:                # best match found
                    colors_here.add(v.color)
                    BMG.add_edge(u.ID, v.ID)            # insert edge (u,v)
            remaining -= colors_here                    # update remaining colors
            parent = parent.parent
    
    # ---- build RBMG as symmetric part of the BMG ----
    for x, neighbors in BMG.adjacency():
        for y in neighbors:
            if BMG.has_edge(y,x):
                RBMG.add_edge(x,y)
                
    return BMG, RBMG


def build_cBMG_alternative(tree):
    """Create an n-colored BMG from a given tree.
    
    Algorithm presented in the cobmg paper."""
    
    V = []
    L = []
    colors = set()
    for v in tree.postorder():
        V.append(v)
        if not v.children:
            L.append(v)
            colors.add(v.color)
    
    l_matrix = np.zeros((len(V), len(colors)), dtype=np.int8)
    v_index = {V[i]: i for i in range(len(V))}
    col_index = {color: i for i, color in enumerate(colors)}
    
    G = nx.DiGraph()
    
    for v in L:
        v.leaves = [v]
        l_matrix[v_index[v], col_index[v.color]] = 1
        G.add_node(v.ID, label=v.label, color=v.color)
    
    for v in tree.postorder():
        if not v.children:
            continue                                        # only traverse inner vertices
        
        for u1, u2 in itertools.combinations(v.children, 2):
            for x in u1.leaves:
                for y in u2.leaves:
                    if l_matrix[v_index[u1], col_index[y.color]] == 0:
                        G.add_edge(x.ID, y.ID)                  # edge (x,y)
                    if l_matrix[v_index[u2], col_index[x.color]] == 0:
                        G.add_edge(y.ID, x.ID)                  # edge (y,x)
        v.leaves = []
        v_nr = v_index[v]
        for u in v.children:
            v.leaves.extend(u.leaves)                       # compute leaves L(v) on the fly
            u_nr = v_index[u]
            for r_nr in range(len(colors)):
                if l_matrix[u_nr, r_nr] == 1:
                    l_matrix[v_nr, r_nr] = 1
    return G


def RBMG_from_BMG(cBMG):
    
    G = nx.Graph()
    G.add_nodes_from(cBMG.nodes(data=True))
    
    for x, neighbors in cBMG.adjacency():
        for y in neighbors:
            if cBMG.has_edge(y,x):
                G.add_edge(x,y)
    return G


# -*- coding: utf-8 -*-

import time, os
import numpy as np

from asymmetree.tools import FileIO, GraphTools
import asymmetree.simulator.TreeSimulator as ts
import asymmetree.simulator.TreeImbalancer as tm
from asymmetree.simulator.Scenario import Scenario
from asymmetree.simulator import NoisyMatrix

from asymmetree.best_matches import TrueBMG
import asymmetree.best_matches.ExtBestHits as ebh
import asymmetree.best_matches.TreeReconstruction as tr
import asymmetree.best_matches.Quartets as qd


class CustomAnalysis:
    
    def __init__(self, scenario, D, epsilon=0.000_000_01):
        
        self.scenario = scenario
        self.D = D
        self.epsilon=epsilon
        

    def compute_statistics(self):
        
        matrix_filename = "temp.phylip"
        species_filename = "temp_species.txt"
        tree_filename = "temp_tree.txt"
        FileIO.matrix_to_phylip(matrix_filename, self.scenario.genes, self.D)
        FileIO.species_to_genes(species_filename, self.scenario)
        FileIO.write_newick(tree_filename, self.scenario.S)
        
        # ---- true BMG, RBMG (based on l.c.a.) ----
        BMG, RBMG = self.scenario.BMG_subtrees, self.scenario.RBMG_subtrees
        self.possible_edges_BMG = self.scenario.possible_edges_BMG()
        
        # ---- epsilon method ----
        start = time.time()
        BMG_eps, RBMG_eps, self.time_eps = ebh.ebh_qinfer(self.scenario, matrix_filename,
                                                          species_filename, self.epsilon)
        BMG_eps, RBMG_eps = self.scenario.reduce_to_subtrees(BMG_eps, RBMG_eps)
        self.BMG_eps_stats = GraphTools.performance(BMG, BMG_eps)
        self.RBMG_eps_stats = GraphTools.performance(RBMG, RBMG_eps)
        
        # ---- neighbor-joining + midpoint rooting ----
        start = time.time()
        nj_tree = tr.neighbor_joining(self.scenario.genes, self.scenario.gene_index, matrix_filename)
        tr.midpoint_rooting(nj_tree)
        BMG_nj, RBMG_nj = TrueBMG.best_match_graphs(nj_tree)
        self.time_nj = time.time() - start
        BMG_nj, RBMG_nj = self.scenario.reduce_to_subtrees(BMG_nj, RBMG_nj)
        self.BMG_nj_stats = GraphTools.performance(BMG, BMG_nj)
        self.RBMG_nj_stats = GraphTools.performance(RBMG, RBMG_nj)
        
        # ---- quadruple method (root subtree outgroups)----
        BMG_qd, RBMG_qd, self.time_qd = qd.quartet_qinfer(self.scenario, matrix_filename,
                                                          species_filename, tree_filename)
        self.BMG_qd_stats = GraphTools.performance(BMG, BMG_qd)
        self.RBMG_qd_stats = GraphTools.performance(RBMG, RBMG_qd)
        
        # ---- quadruple method (all outgroups)----
        BMG_qd2, RBMG_qd2, self.time_qd2 = qd.quartet_qinfer(self.scenario, matrix_filename,
                                                             species_filename, tree_filename,
                                                             closest_outgroups=True,
                                                             incongruence_threshold=0.2)
        self.BMG_qd2_stats = GraphTools.performance(BMG, BMG_qd2)
        self.RBMG_qd2_stats = GraphTools.performance(RBMG, RBMG_qd2)
        
        os.remove(matrix_filename)
        os.remove(species_filename)
        os.remove(tree_filename)

    
    def summary_line(self):
        data = [*self.scenario.rates_and_counts(),
                self.epsilon,
                self.scenario.BMG_subtrees.size(), self.scenario.RBMG_subtrees.size(),
                self.possible_edges_BMG,
                *self.BMG_eps_stats, *self.RBMG_eps_stats,
                *self.BMG_nj_stats, *self.RBMG_nj_stats,
                *self.BMG_qd_stats, *self.RBMG_qd_stats,
                *self.BMG_qd2_stats, *self.RBMG_qd2_stats,
                self.time_eps, self.time_nj, self.time_qd, self.time_qd2]
        
        line = "{}" + (len(data)-1) * "\t{}"
        line = line.format(*data)
        return line
    
    
    def write_summary_line(self, filename, *first_columns):
        line = "\n" + (len(first_columns)) * "{}\t"
        line = line.format(*first_columns) + self.summary_line()
        with open(filename, "a") as f:
            f.write(line)
        
    
    @staticmethod
    def summary_header():
        data = ["D_rate", "L_rate", "H_rate",
                "S_count", "D_count", "L_count", "H_count", "ancient_dupl",
                "epsilon",
                "BMG_size", "RBMG_size",
                "possible_edges_BMG"]
        for mode in ("eps", "nj", "qd", "qd2"):
            for graph_type in ("BMG", "RBMG"):
                for stat in ("order", "size", "tp", "tn", "fp", "fn", "acc", "prec", "recall"):
                    data.append(graph_type + "_" + mode + "_" + stat)
        data.extend(["time_eps", "time_nj", "time_qd", "time_qd2"])
        
        header = "{}" + (len(data)-1) * "\t{}"
        header = header.format(*data)
        return header
    
    
    @staticmethod
    def write_header(filename, *first_columns):
        line = (len(first_columns)) * "{}\t"
        line = line.format(*first_columns) + CustomAnalysis.summary_header()
        with open(filename, "w") as f:
            f.write(line)


# ---- simulation parameters ----

S_min, S_max = 3, 20            # range number of species
D_min, D_max = 0.5, 1.0         # range of dupl. rate
L_min, L_max = 0.5, 1.0         # range of loss rate
H_min, H_max = 0.0, 0.0         # range of HGT rate

repeats = 10
epsilon = 0.5

noise_sds = [0.0, 0.05, 0.1, 0.2, 0.3, 0.5, 0.75, 1.0, 1.5]
noise_alphas = []#[0.0, 0.02, 0.05, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6]

sim_ID = 0

stats_file = "results/method_comparison.csv"

# write header of the main statistics file
CustomAnalysis.write_header(stats_file, "SCEN_ID", "noise_type", "sd", "true_sd", "alpha")


for rep in range(repeats):
    
    D = np.random.uniform(low=D_min, high=D_max) if D_max > D_min else 0.0
    L = np.random.uniform(low=L_min, high=L_max) if L_max > L_min else 0.0
    H = np.random.uniform(low=H_min, high=H_max) if H_max > H_min else 0.0
    
    S = ts.build_species_tree(np.random.randint(S_min, S_max+1), planted=True)
    
    TGT_simulator = ts.GeneTreeSimulator(S)
    TGT1 = TGT_simulator.simulate((D,L,H))
    TGT1 = tm.imbalance_tree(TGT1, S, baseline_rate=1,
                            lognormal_v=0.2,
                            gamma_param=(0.5, 1.0, 2.2),
                            weights=(1/3, 1/3, 1/3))
    
    TGT2 = TGT_simulator.simulate((D,L,H))
    TGT2 = tm.imbalance_tree(TGT2, S, baseline_rate=1,
                            lognormal_v=0.2,
                            gamma_param=(0.5, 1.0, 2.2),
                            weights=(1/3, 1/3, 1/3))
    
    scenario1 = Scenario(S, TGT1, (D,L,H))
    scenario2 = Scenario(S, TGT2, (D,L,H))
    
    D1 = scenario1.get_distance_matrix()
    D2 = scenario2.get_distance_matrix()
    
    for noise_sd in noise_sds:
        D1_noisy = NoisyMatrix.noisy_matrix(D1, noise_sd)
        D2_noisy = NoisyMatrix.noisy_matrix(D2, noise_sd)
        true_sd1 = np.nanstd(np.divide(D1_noisy, D1)) if noise_sd > 0.0 else 0.0
        true_sd2 = np.nanstd(np.divide(D2_noisy, D2)) if noise_sd > 0.0 else 0.0
        
        summary1 = CustomAnalysis(scenario1, D1_noisy, epsilon=epsilon)
        summary1.compute_statistics()
        summary1.write_summary_line(stats_file, sim_ID, "no_bias", noise_sd, true_sd1, 0.0)
        
        summary2 = CustomAnalysis(scenario2, D2_noisy, epsilon=epsilon)
        summary2.compute_statistics()
        summary2.write_summary_line(stats_file, sim_ID+1, "no_bias", noise_sd, true_sd2, 0.0)
        
        print(sim_ID/ (2*repeats), "Noise:", noise_sd)
    
    for noise_alpha in noise_alphas:
        D1_noisy, D2_noisy = NoisyMatrix.convex_linear_comb(D1, D2, alpha=noise_alpha)
        true_sd1 = np.nanstd(np.divide(D1_noisy, D1)) if noise_alpha > 0.0 else 0.0
        true_sd2 = np.nanstd(np.divide(D2_noisy, D2)) if noise_alpha > 0.0 else 0.0
        
        summary1 = CustomAnalysis(scenario1, D1_noisy, epsilon=epsilon)
        summary1.compute_statistics()
        summary1.write_summary_line(stats_file, sim_ID, "bias", 0.0, true_sd1, noise_alpha)
        
        summary2 = CustomAnalysis(scenario2, D2_noisy, epsilon=epsilon)
        summary2.compute_statistics()
        summary2.write_summary_line(stats_file, sim_ID+1, "bias", 0.0, true_sd2, noise_alpha)
        
        print(sim_ID/ (2*repeats), "alpha:", noise_alpha)
    
    sim_ID += 2
    
    print(sim_ID / (2*repeats))


# -*- coding: utf-8 -*-

import collections, itertools, random

from asymmetree.tools import Tree


__author__ = "David Schaller"
__copyright__ = "Copyright (C) 2019, David Schaller"


class SimpleGraph:
    
    def __init__(self, initial=None):
        if isinstance(initial, dict):
            self.adj_list = initial
        elif isinstance(initial, collections.Iterable):
            self.adj_list = {v: set() for v in initial}
        else:
            self.adj_list = {}
    
    
    def add_node(self, v):
        if v not in self.adj_list:
            self.adj_list[v] = set()
    
    
    def has_edge(self, u, v):
        if u in self.adj_list and v in self.adj_list[u]:
            return True
        else:
            return False
        
    
    def neighbors(self, v):
        return iter(self.adj_list[v])
    
    
    def add_edge(self, u, v):
        if u == v:
            raise ValueError("No loops allowed!")
        # add nodes if not yet in the graph
        self.add_node(u)
        self.add_node(v)
        
        self.adj_list[u].add(v)
        self.adj_list[v].add(u)
    
    
    def connected_comp(self, make_undirected=False):
        """Determines connected components by breadth-first search.
        
        Return a list of dictionaries.
        """
        result = []                                             # a list of dictionaries
        visited = set()                                         # already visited nodes
        Q = collections.deque()                                 # initialize queue
        for start_node in self.adj_list.keys():
            if start_node in visited:
                continue
            conn_comp = dict()                                  # dictionary that represents a c.c.
            conn_comp[start_node] = self.adj_list[start_node]
            visited.add(start_node)
            Q.append(start_node)
            while Q:
                v = Q[0]
                for u in self.adj_list[v]:
                    if u not in visited:
                        visited.add(u)
                        Q.append(u)
                        conn_comp[u] = self.adj_list[u]
                Q.popleft()
            result.append(conn_comp)
        return result
    
    
    def graphs_equal(self, other):
        """Returns whether two undirected graphs are equal."""
        
        set1 = set(self.adj_list.keys())
        set2 = set(other.adj_list.keys())
        
        if set1 != set2:
            return False
        
        for v in set1:
            if self.adj_list[v] != other.adj_list[v]:
                return False
        
        return True
    
    
    def symmetric_diff(self, other):
        """Returns the number of edges in the symmetric difference."""
        
        set1 = set(self.adj_list.keys())
        set2 = set(other.adj_list.keys())
        
        if set1 != set2:
            raise ValueError("Graphs do not have the same vertex set")
            return
        
        V = list(set1)
        sym_diff_number = 0
        
        for i in range(0, len(V)-1):
            for j in range(i+1, len(V)):
                if self.has_edge(V[i], V[j]) and not other.has_edge(V[i], V[j]):
                    sym_diff_number += 1
                elif not self.has_edge(V[i], V[j]) and other.has_edge(V[i], V[j]):
                    sym_diff_number += 1
        
        return sym_diff_number
    
    
    def get_complement(self):
        
        compl_graph = SimpleGraph(initial=self.adj_list.keys())
        
        V = list(self.adj_list.keys())
        
        for i in range(len(V)-1):
            for j in range(i+1, len(V)):
                if not self.has_edge(V[i], V[j]):
                    compl_graph.add_edge(V[i], V[j])
        
        return compl_graph
    
    
    @staticmethod
    def random_graph(N, p=0.5):
        """Construct a random graph on N nodes.
        
        Keyword arguments:
            p - probability that an edge xy is inserted"""
        V = [i for i in range(1, N+1)]
        
        G = SimpleGraph(initial=V)
        
        for i in range(0, len(V)-1):
            for j in range(i+1, len(V)):
                if random.random() < p:
                    G.add_edge(V[i], V[j])
        
        return G
    

class CotreeNode(Tree.TreeNode):
    
    __slots__ = ['aux_counter']
    
    
    def __init__(self, ID, label=None, parent=None,
                 aux_counter=0):
        
        super().__init__(ID, label=label)
        
        self.aux_counter = aux_counter      # auxiliary counting variable
        
    
    def __str__(self):
        if not self.children:
            return str(self.ID)
        elif self.label == "series":
            return "<1>"
        elif self.label == "parallel":
            return "<0>"
        else:
            return "<>"
        

class Cotree(Tree.Tree):
    
    # corresponding node type
    node_type = CotreeNode
    
    
    def __init__(self, root):
        super().__init__(root)
        
    
    def to_cograph(self):
        self.supply_leaves()
        G = SimpleGraph()
        
        for v in self.root.leaves:
            G.add_node(v.ID)
        
        for u in self.preorder():
            if u.label == "series":
                for child1, child2 in itertools.combinations(u.children, 2):
                    for v1 in child1.leaves:
                        for v2 in child2.leaves:
                            G.add_edge(v1.ID, v2.ID)
        
        return G
        
        
    @staticmethod
    def cotree(G):
        """Checks if a graph is a cograph and returns its cotree.
        
        Simple O(n^3) implementation."""
        
        def build_cotree(G, label=None):
            v = CotreeNode(None)
            v.label = label
            child_nodes = []
            
            if len(G.adj_list) == 1:
                v.label = "leaf"
                for ID in G.adj_list.keys():
                    v.ID = ID
            
            elif v.label is None:
                ccs = G.connected_comp()
                if len(ccs) > 1:
                    v.label = "parallel"
                    
                    for cc in ccs:
                        G_i = SimpleGraph(initial=cc)
                        v_i = build_cotree(G_i, label="series")
                        if v_i:
                            child_nodes.append(v_i)
                        else:
                            return False
                else:
                    G = G.get_complement()
                    ccs = G.connected_comp()
                    if len(ccs) > 1:
                        v.label = "series"
                        
                        for cc in ccs:
                            G_i = SimpleGraph(initial=cc)
                            v_i = build_cotree(G_i, label="parallel")
                            if v_i:
                                child_nodes.append(v_i)
                            else:
                                return False
                    else:
                        return False
            
            else:
                G = G.get_complement()
                ccs = G.connected_comp()
                if len(ccs) > 1:                 
                    for cc in ccs:
                        child_label = "series" if v.label=="parallel" else "parallel"
                        G_i = SimpleGraph(initial=cc)
                        v_i = build_cotree(G_i, label=child_label)
                        if v_i:
                            child_nodes.append(v_i)
                        else:
                            return False
                else:
                    return False
            
            for child in child_nodes:
                v.children.append(child)
                child.parent = v
            
            return v
        
        root = build_cotree(G)
        
        if root:
            return Cotree(root)
        else:
            print("Not a Cograph!")
            return False
       
    
    @staticmethod
    def random_cotree(N_leaves, force_series_root=False):
        """Creates a random cotree."""
        
        root = CotreeNode(ID=0)
        cotree = Cotree(root)
        node_list = [root]
        nr, leaf_count = 1, 1
        
        while leaf_count < N_leaves:
            node = random.choice(node_list)
            if not node.children:                       # avoid nodes with outdegree 1
                new_child1 = CotreeNode(nr)
                new_child2 = CotreeNode(nr+1)
                node.add_child(new_child1)
                node.add_child(new_child2)
                node_list.extend(node.children)
                nr += 2
                leaf_count += 1
            elif node.children:                         # add only one child if there are already children
                new_child = CotreeNode(nr)
                node.add_child(new_child)
                node_list.append(new_child)
                nr += 1
                leaf_count += 1
                
        for v in cotree.preorder():                     # assign labels ('series', 'parallel', 'leaf')
            if not v.children:
                v.label == "leaf"
            elif v.parent is None:
                if force_series_root:
                    v.label = "series"
                else:
                    v.label = "series" if random.random() < 0.5 else "parallel"
            else:
                v.label = "series" if v.parent.label == "parallel" else "parallel"
                
        return cotree
    
    
if __name__ == "__main__":
    
    from pprint import pprint
    
    cotree = Cotree.random_cotree(100)
    print(cotree.to_newick())
    cograph = cotree.to_cograph()
    #pprint(cograph.adj_list)
    
    cotree2 = Cotree.cotree(cograph)
    print(cotree2.to_newick())
    cograph2 = cotree2.to_cograph()
    #pprint(cograph2.adj_list)
    
    print(cograph.graphs_equal(cograph2))
    
    cograph3 = SimpleGraph()
    cograph3.add_edge("a", "b")
    cograph3.add_edge("b", "d")
    cograph3.add_edge("c", "d")
    cograph3.add_edge("b", "c")
    cotree3 = Cotree.cotree(cograph3)
    print(cotree3.to_newick())
    
    cograph4 = SimpleGraph.random_graph(10, p=0.5)
    pprint(cograph4.adj_list)
    cotree4 = Cotree.cotree(cograph4)
    if cotree4:
        print(cotree4.to_newick())


# -*- coding: utf-8 -*-

import os, pickle, random

import pyvolve as pv

import asymmetree.simulator.TreeSimulator as ts
import asymmetree.simulator.TreeImbalancer as tm


def to_newick_pyvolve(tree, node=None):
    
    if node is None:
        return to_newick_pyvolve(tree, node=tree.root) + ";"
    elif not node.children:
        return "{}_{}:{}".format(node.color, node.label, node.dist)
    else:
        s = ''
        for child in node.children:
            s += to_newick_pyvolve(tree, node=child) + ","
        return "({}):{}".format(s[:-1], node.dist)
    
    
def to_newick_faa(tree, node=None):
    
    if node is None:
        return to_newick_faa(tree, node=tree.root) + ";"
    elif not node.children:
        return "{}.faa".format(node.label)
    else:
        s = ''
        for child in node.children:
            s += to_newick_faa(tree, node=child) + ","
        return "({})".format(s[:-1])


def simulate_gene_families(S, N, seq_length=(200,800),
                           scaling_factor=1.0):
    
    seq_dict = {}
    gene_trees = []
    
    for v in S.preorder():
        if not v.children:
            seq_dict[str(v.label)] = []
            
    for i in range(N):                         
        TGT_simulator = ts.GeneTreeSimulator(S)
        TGT = TGT_simulator.simulate((1.0, 0.5, 0.0))       # only dupl./loss, HGT is disabled
        TGT = tm.imbalance_tree(TGT, S, baseline_rate=1,
                                lognormal_v=0.2,
                                gamma_param=(0.5, 1.0, 2.2),
                                weights=(1, 1, 1))
        OGT = ts.observable_tree(TGT)
        
        for v in OGT.preorder():
            v.dist *= scaling_factor
        
        phylogeny = pv.read_tree(tree=to_newick_pyvolve(OGT))
        model = pv.Model("JTT")
        
        seq_size = random.randint(*seq_length) if isinstance(seq_length, tuple) else seq_length
        partition = pv.Partition(models=model, size=seq_size)
        evolver = pv.Evolver(tree=phylogeny, partitions=partition)
        evolver(seqfile=False, ratefile=False, infofile=False)
        for label, seq in evolver.get_sequences().items():
            species, gene_label = label.split("_", 1)
            seq_dict[species].append( (i, gene_label, seq) )
        
        gene_trees.append( (i, OGT) )
    
    return seq_dict, gene_trees


def write_species_fastas(seq_dict, dirname, pickle_scenario=None):
        
    for species, seq_list in seq_dict.items():
        
        filename = os.path.join(dirname, "{}.faa".format(species))
        with open(filename, "w") as f:
            for fam_id, gene_label, seq in seq_list:
                f.write(">fam{}_gene{}\n".format(fam_id, gene_label))
                pos = 0
                while pos < len(seq):
                    f.write(seq[pos:min(pos+80, len(seq))])
                    pos += 80
                    f.write("\n")
    
    if pickle_scenario:
        pickle_file = os.path.join(dirname, "scenario.pickle")
        pickle.dump(pickle_scenario, open(pickle_file , "wb" ))


# -*- coding: utf-8 -*-

""" 
Implementation of the BuildST algorithm described by Deng & Fernández-Baca.

Builds a supertree from a profile of phylogenetic trees. Uses the algorithm
described by Deng and Fernández-Baca in 2016. The class 'BuildST' accepts a 
list of trees and computes the supertree if it exists.


References:
    - Yun Deng and David Fernández-Baca. Fast Compatibility Testing for
      Rooted Phylogenetic Trees. 27th Annual Symposium on Combinatorial
      Pattern Matching (CPM 2016). DOI: 10.4230/LIPIcs.CPM.2016.12
"""

from collections import deque

from asymmetree.tools.Tree import Tree, TreeNode
from asymmetree.tools.DoublyLinkedList import DLList
from asymmetree.tools.hdtgraph.DynamicGraph import Graph as HDTGraph


__author__ = "David Schaller"
__copyright__ = "Copyright (C) 2018, David Schaller"


class BuildST:
    """Deng and Fernández-Baca's BuildST algorithm."""
    
    def __init__(self, treelist, hdt=True):
        """Constructor for BuildST algorithm."""
        
        self.treelist = treelist
        self.Xp = dict()                                # maps leaf label --> XpNode
        self.marked = set()                             # set of marked nodes
        self.node_to_tree_index = dict()
        self.list_pointer = dict()                      # maps node --> DLL element
        self.singleton_pointer = dict()                 # maps node --> DLL element
        
        self.id_counter = 0                             # counter for tree nodes
        
        if hdt:
            self.ConnectedComp = ConnectedComp_HDT      # use class 'ConnectedComp_HDT'
            self.hdt_graph = HDTGraph()
        else:
            self.ConnectedComp = ConnectedComp_simple   # use class 'ConnectedComp_simple'
            self.hdt_graph = False
        
    
    def run(self):
        """Build the supertree from the given tree list if existent."""
        
        self._prepare_trees()        # tree indices, Xp nodes
        U_init = self._initialize()
        if not U_init:
            return False
        root = self._buildst(U_init)
        if not root:
            print("Could not build the supertree!")
            return False
        else:
            return Tree(root)
    
    
    def _prepare_trees(self):
        
        for i in range(len(self.treelist)):
            tree = self.treelist[i]
            for node in tree.preorder():
                # access to tree index i in [k]
                self.node_to_tree_index[node] = i
                if not node.children:
                    if node.label not in self.Xp:
                        self.Xp[node.label] = XpNode(node.label)
                    self.Xp[node.label].leafnodes.append(node)
    
    
    def _initialize(self):
        
        Y = self.ConnectedComp(len(self.treelist), self, count=len(self.Xp))
        
        for i in range(len(self.treelist)):
            
            tree = self.treelist[i]
            
            # Y_init.singleton is the set [k]
            self.singleton_pointer[tree.root] = Y.singleton.append(i)
            
            # append the root to DLList at Y.List[i]
            self.list_pointer[tree.root] = Y.List[i].append(tree.root)
            
            Y.initialize_tree_edges(tree)
            
            # mark root as part of set U_init
            self.marked.add(tree.root)
                          
        # glue together the leafnodes and the labelnodes in Xp
        for labelnode in self.Xp.values():
            for leafnode in labelnode.leafnodes:
                Y.add_edge(labelnode, leafnode)
                
        if self.hdt_graph:
            ett = self.hdt_graph.is_connected()
            if ett:
                Y.representative = self.treelist[0].root
            else:
                print("Initialization failed. Graph is not fully connected!")
                return
            
        return Y
    
    
    def _buildst(self, U):
        
        r_U = TreeNode(self.id_counter)    # create a node r_U
        self.id_counter += 1
        
        # --------------------------------------------------
        # if |L(U)| = 1 then
        #    return the tree consisting of node r_U,
        #    labeled by the single species in L(U)
        # --------------------------------------------------
        if U.count == 1:
            label = U.get_labelnodes()[0]
            r_U.label = label.label
            return r_U
        
        # --------------------------------------------------
        # if |L(U)| = 2 then
        #    return the tree consisting of node r_U and two children
        #    each labeled by a different species in L(U)
        # --------------------------------------------------
        if U.count == 2:
            labels = U.get_labelnodes()
            if len(labels) != 2:
                print("Could not find 2 label nodes!")
                return
            node1 = TreeNode(self.id_counter, label=labels[0].label)
            node2 = TreeNode(self.id_counter+1, label=labels[1].label)
            self.id_counter += 2
            r_U.add_child(node1)
            r_U.add_child(node2)
            return r_U
        
        # --------------------------------------------------
        # foreach i in [k] such that |U(i)| = 1 do
        #    Let v be the single element in U(i)
        #    U = (U \ {v}) union Ch(v)
        # --------------------------------------------------
        W = [U]                             # list of conn. components
        J = [j for j in U.singleton]        # elements in singleton
        V = [U.List[i][0] for i in J]       # singleton nodes to be deleted
        
        for i in J:
            v_i = U.List[i][0]
            U.List[i].remove_element(self.list_pointer[v_i])
            U.singleton.remove_element(self.singleton_pointer[v_i])
            
            self.marked.discard(v_i)        # unmark v_i and
            for child in v_i.children:      # mark its children
                self.marked.add(child)
                self.list_pointer[child] = U.List[i].append(child)
        
        for v_i in V:
            Y_current = None
            
            # which component in W contains v_i?
            for Y in W:
                if Y.contains_key(v_i):
                    Y_current = Y
                    break
                
            if not Y_current:
                print("Could not find connected component for v_i", v_i)
                return
            
            for u in v_i.children:
                conn_comp = Y_current.delete_edge(v_i, u)
                if len(conn_comp) == 2:
                    if v_i in conn_comp[0]:
                        
                        # (case 1.1) u is the last child of v_i
                        #            then conn_comp[0] is actually empty
                        if len(conn_comp[0]) == 1:
                            Y_current.set_component(component=conn_comp[1], representative=u)
                            break
                        
                        # (case 1.2) two actual components emerge
                        else:
                            Y_current.set_component(component=conn_comp[0], representative=v_i)
                            Y_new = self.ConnectedComp(Y_current.k, self)
                            Y_new.set_component(component=conn_comp[1], representative=u)
                            current_smaller = True if len(conn_comp[0]) < len(conn_comp[1]) else False
                    
                    elif v_i in conn_comp[1]:
                        
                        # (case 2.1) u is the last child of v_i
                        #            then conn_comp[1] is actually empty
                        if len(conn_comp[1]) == 1:
                            Y_current.set_component(component=conn_comp[0], representative=u)
                            break
                        
                        # (case 2.2) two actual components emerge
                        else:
                            Y_current.set_component(component=conn_comp[1], representative=v_i)
                            Y_new = self.ConnectedComp(Y_current.k, self)
                            Y_new.set_component(component=conn_comp[0], representative=u)
                            current_smaller = True if len(conn_comp[1]) < len(conn_comp[0]) else False
                    
                    if current_smaller:
                        self._update_conn_comps(Y_current, Y_new, first_swap=True)
                    else:
                        self._update_conn_comps(Y_new, Y_current)
                    
                    W.append(Y_new)
                    
        # --------------------------------------------------
        # Let W_1, W_2, ..., W_p be the connected components
        # if p = 1 then
        #    return incompatible
        # --------------------------------------------------
        if len(W) == 1:
            print("Leaves:", W[0].get_labelnodes() )
            print("The tree are incompatible!")
            return False
        
        # --------------------------------------------------
        # foreach j in [p] do
        #    Let t_j = BuildST(W_j)
        #    if t j is a tree then
        #       Add t_j to the set of subtrees of r_U
        #    else
        #       return incompatible
        # --------------------------------------------------
        for W_j in W:
            t_j = self._buildst(W_j)
            if t_j:
                r_U.add_child(t_j)
            else:
                return False
        
        # return the tree with root r_U
        return r_U
    
    
    def _update_conn_comps(self, Y1, Y2, first_swap=False):
        """Updates Y.count, Y.singleton and Y.List of two newly split components.
        
        Component Y1 is supposed to be the smaller one and is scanned for label
        nodes (Xp) and for marked nodes.
        Keyword argument:
            first_swap - if Y1 (smaller component) contains the information
                         about count, List and singleton instead of Y2
                         these attributes are swapped
        """
        
        if first_swap:
            Y1.count, Y2.count = Y2.count, Y1.count
            Y1.List, Y2.List = Y2.List, Y1.List
            Y1.singleton, Y2.singleton = Y2.singleton, Y1.singleton
            
        for v in Y1.keys():
            if isinstance(v, XpNode):
                Y1.count += 1
                Y2.count -= 1
            elif v in self.marked:
                i = self.node_to_tree_index[v]
                Y2.List[i].remove_element(self.list_pointer[v])
                self.list_pointer[v] = Y1.List[i].append(v)
                
                # v becomes singleton
                if len(Y2.List[i]) == 1:
                    singleton_node = Y2.List[i][0]
                    self.singleton_pointer[singleton_node] = Y2.singleton.append(i)
                # index i is no longer singleton, i.e. it must have been a singleton
                elif len(Y2.List[i]) == 0:
                    Y2.singleton.remove_element(self.singleton_pointer[v])
                
                # v becomes singleton
                if len(Y1.List[i]) == 1:                                        
                    self.singleton_pointer[v] = Y1.singleton.append(i)
                # index i is no longer singleton
                elif len(Y1.List[i]) == 2:
                    Y1.singleton.remove_element(self.singleton_pointer[Y1.List[i][0]])


class XpNode:
    """Special type of node for the set Xp (one for each leaf label).""" 
    
    def __init__(self, label):
        
        self.label = label
        self.leafnodes = []         # corresponding treenodes of the profile
    
    def __repr__(self):
        
        return "<XpNodeID:" + str(id(self)) + ", " + str(self.label)
    
    def __str__(self):
        
        return str(self.label)
        
    
class ConnectedComp_HDT:
    """Connected components class for D.-F.-B. algorithm (HDT datastructure)."""
    
    def __init__(self, k, buildst, count=0, singleton=None, List=None):
        
        self.k = k
        self.buildst = buildst
        self.hdt_graph = buildst.hdt_graph
        self.count = count
        self.singleton = singleton if singleton else DLList()
        self.List = List if List else [DLList() for i in range(k)]
        
        # to find the component in the HDT datastructure
        self.representative = None
    
    
    def keys(self):
        """Generator for the elements in the component.
        
        Caution: If the HDT datastructure is used this will only work
        if a representative of the component is set correctly.
        """
        
        yield from self.hdt_graph.component_iterator(self.representative)
    
    
    def contains_key(self, key):
        """Check if a given key is element of the component."""
        
        if not self.representative: 
            print("Missing representative for the component! (2)")
            return False
        return self.hdt_graph.connected(self.representative, key)
    
    
    def initialize_tree_edges(self, tree):
        """Add all edges of a tree to the graph datastructure."""
        
        self.hdt_graph.add_loose_tree(tree)
    
    
    def add_edge(self, u, v):
        """Add an edge to the graph datastructure."""
        
        self.hdt_graph.insert_edge(u, v)
    
    
    def delete_edge(self, u, v):
        """Delete an edge and return the 1 or 2 conn. components of u and v."""
        
        self.hdt_graph.delete_edge(u, v)
        ett1 = self.hdt_graph.get_component(u)
        ett2 = self.hdt_graph.get_component(v)
        if ett1 is ett2:
            return [ett1]
        else:
            return [ett1, ett2]
    
    
    def set_component(self, component=None, representative=None):
        """Set the component (self.representative)."""
        
        self.representative = representative
    
    
    def get_labelnodes(self):
        """Return the labelnodes in the connected component."""
        
        result = []
        if not self.representative:
            print("Missing representative for the component! (1)")
            return
        for node in self.hdt_graph.component_iterator(self.representative):
            if isinstance(node, XpNode):
                result.append(node)
                
        return result


class ConnectedComp_simple:
    """Connected components class for D.-F.-B. algorithm (simple datastructure)."""
    
    def __init__(self, k, buildst, count=0, singleton=None, List=None):
        
        self.k = k
        self.buildst = buildst
        self.count = count
        self.singleton = singleton if singleton else DLList()
        self.List = List if List else [DLList() for i in range(k)]
        
        self.component = dict()
    
    
    def keys(self):
        """Generator for the elements in the component."""
        
        yield from self.component.keys()
    
    
    def contains_key(self, key):
        """Check if a given key is element of the component."""
        
        return key in self.component
        
    
    def initialize_tree_edges(self, tree):
        """Add all edges of a tree to the graph datastructure."""
        
        for u, v in tree.edges():
            if u not in self.component:
                self.component[u] = set()
            if v not in self.component:
                self.component[v] = set()
            self.component[u].add(v)         # add all tree edges
            self.component[v].add(u)         # as undirected edges
    
    
    def add_edge(self, u, v):
        """Add an edge to the graph datastructure."""
        
        if u not in self.component:
            self.component[u] = set()
        if v not in self.component:
            self.component[v] = set()
        self.component[u].add(v)             # add all tree edges
        self.component[v].add(u)             # as undirected edges
    
    
    def delete_edge(self, u, v):
        """Delete an edge and return the 1 or 2 conn. components of u and v."""
        
        self.component[u].remove(v)
        self.component[v].remove(u)
        return self.breadth_first_search()     
    
    
    def breadth_first_search(self):
        """Determines connected components by breadth-first search."""
        
        result = []                                 # a list of dictionaries
        visited = set()                             # already visited nodes
        Q = deque()                                 # initialize queue
        
        for start_node in self.component.keys():
            if start_node in visited:
                continue
            conn_comp = dict()                      # dictionary that represents
                                                    # one conn. comp.
            conn_comp[start_node] = self.component[start_node]
            visited.add(start_node)
            Q.append(start_node)
            while Q:
                v = Q[0]
                for u in self.component[v]:
                    if u not in visited:
                        visited.add(u)
                        Q.append(u)
                        conn_comp[u] = self.component[u]
                Q.popleft()
            result.append(conn_comp)
            
        return result
    
    
    def set_component(self, component=None, representative=None):
        """Set the component (self.component)."""
        
        self.component = component
    
    
    def get_labelnodes(self):
        """Return the labelnodes in the connected component."""
        
        return [node for node in self.component.keys() if isinstance(node, XpNode)]


if __name__ == "__main__":
    
    # ----- TESTING THIS MODULE -----
    
    import numpy as np
    from asymmetree.tools.PhyloTree import PhyloTree
    from asymmetree.simulator.TreeSimulator import observable_tree
    
    N = 20
    t = PhyloTree.random_colored_tree(N, [1,2,3], binary=False)
    print(t.to_newick(color=False, distance=False, label_inner=False))
    
    partial_trees = []
    for i in range(10):
        
        mark = np.random.choice(N, size=N//2)
        
        j = 0
        for v in t.preorder():
            if not v.children:
                if j in mark:
                    v.label = "*"
                else:
                    v.label = str(v.ID)
                j += 1
        obs_tree = observable_tree(t)
        partial_trees.append(obs_tree)
    
    # restore labels
    for v in t.preorder():
        v.label = str(v.ID)
    
    st_builder = BuildST(partial_trees)
    supertree = st_builder.run()
        
    print(supertree.to_newick())
    
    print(t.compare_topology(supertree))


# -*- coding: utf-8 -*-

"""
Tree Simulator.

Simulate species and gene trees.

Methods in this module:
    - build_species_tree
    - make_ultrametric
    - build_gene_tree
    - observable_tree
"""

import random
from collections import deque

import numpy as np

from asymmetree.tools.PhyloTree import PhyloTree, PhyloTreeNode


__author__ = "David Schaller"
__copyright__ = "Copyright (C) 2020, David Schaller"


# --------------------------------------------------------------------------
#                    CONSTRUCTION OF THE SPECIES TREE
#
#                       with the innovations model
# --------------------------------------------------------------------------

def build_species_tree(N, planted=True, model="innovations",
                       non_binary=0.0):
    """Builds a species tree S with N leaves.
    
    Keyword parameters:
        planted -- add a planted root that has the canonical root as its
                   single neighbor; default is True
        model -- simulation model to be applied; default is 'innovations'
        non_binary -- probability that an inner edge is contracted;
                      results in non-binary tree; default is 0.0
    """
    
    if isinstance(model, str) and model.lower() in ('innovation', 'innovations'):
        tree = _innovations_model(N, planted=planted)
    else:
        raise ValueError("Model '{}' is not available!".format(model))
        
    if non_binary > 0.0:
         edges = _select_edges_for_contraction(tree,
                                               min(non_binary, 1.0),
                                               exclude_planted_edge=True)
         tree.contract(edges)
         
    make_ultrametric(tree)
        
    return tree
        

def _innovations_model(N, planted=True):
    """Builds a species tree S with N leaves with the innovations model."""
    
    tree = PhyloTree(PhyloTreeNode(0, label="0"))
    tree.number_of_species = N
    
    if planted:                     # planted tree (root is an implicit
                                    # outgroup with outdegree = 1)
        root = PhyloTreeNode(1, label="1")
        tree.root.add_child(root)
        node_counter = 2
    else:
        root = tree.root
        node_counter = 1
    
    t = 1
    features = [0]                  # set of available features
    species = {(0,): root}          # extant species
    
    while len(species) < N:
        
        loss_candidates = set()     # species for which loss of a feature
        for s in species.keys():    # can trigger a speciation
            for i in range(0, len(s)):
                    if s[:i] + s[i+1:] not in species:
                        loss_candidates.add(s)
        
        if not loss_candidates:     # INNOVATION EVENT
            s = random.choice(list(species))
            new_feature = len(features)
            
            new_s = s + (new_feature,)
            
            child1 = PhyloTreeNode(node_counter, label=str(node_counter))
            species[s].add_child(child1)
            child2 = PhyloTreeNode(node_counter+1, label=str(node_counter+1))
            species[s].add_child(child2)
            
            node_counter += 2
            
            species[s] = child1
            species[new_s] = child2
            features.append(new_feature)
            t += 1
            
        else:
            s = random.choice(list(loss_candidates))
            
            if len(s) > 1:
                feature_index = random.randint(0, len(s)-1)
            else:
                feature_index = 0
            
            new_s = s[:feature_index] + s[feature_index+1:]
            
            if new_s not in species:    # LOSS EVENT
                
                child1 = PhyloTreeNode(node_counter, label=str(node_counter))
                species[s].add_child(child1)
                child2 = PhyloTreeNode(node_counter+1, label=str(node_counter+1))
                species[s].add_child(child2)

                node_counter += 2
                
                species[s] = child1
                species[new_s] = child2
                t += 1
    
    return tree


def make_ultrametric(tree):
    """Makes a given species tree S ultrametric.
    
    It is t(root) = 1 and t(x) = 0 for x in L(S)."""
    for v in tree.preorder():
        if not v.parent:
            v.dist = 0.0
            v.tstamp = 1.0
        elif not v.children:
            v.dist = v.parent.tstamp
            v.tstamp = 0.0
        else:                               # random walk to a leaf
            pos = v                         # current position
            length = 0                      # path length |P|
            while pos.children:
                length += 1
                pos = pos.children[np.random.randint(len(pos.children))]
            v.dist = (v.parent.tstamp) * 2 * np.random.uniform() / (length+1)
            v.tstamp = v.parent.tstamp - v.dist
            
            
def _select_edges_for_contraction(tree, p, exclude_planted_edge=True):
    
    edges = []
    
    for u, v in tree.inner_edges():
        
        if exclude_planted_edge and (u is tree.root) and len(u.children) == 1:
            continue
        
        if random.random() < p:
            edges.append((u,v))
    
    return edges


# --------------------------------------------------------------------------
#                    CONSTRUCTION OF THE GENE TREE
#
# --------------------------------------------------------------------------


class Branch:
    
    def __init__(self, ID, parent, S_u, S_v, transferred):
        
        self.ID = ID
        self.parent = parent
        self.S_u = S_u
        self.S_v = S_v
        self.transferred = transferred
        

class GeneTreeSimulator:
    
    def __init__(self, S):
        
        self.S = S                                      # species tree
        self.sorted_speciations = S.sorted_nodes()      # list of speciations sorted by time stamp
        self.sorted_edges = S.sorted_edges()            # edges of the species tree
                                                        # sort (u,v) by tstamp of u
    
    
    def simulate(self, DLH_rates,
                 dupl_polytomy=0.0):
        """Simulate a gene tree along the specified species tree.
        
        Simulate a gene tree along the specified species tree where the parameter
        'DLH_rates' contains the rates for duplications, losses and HGTs.
        
        Keyword arguments:
            dupl_polytomy - allows non-binary duplication events by specifying
                            the lambda parameter for a poisson distribution
                            (copy number = drawn number + 2); default is 0.0
        """
        
        self.DLH_rates = DLH_rates
        self.rate_sum = sum(DLH_rates)
        self.d = DLH_rates[0]
        self.l = DLH_rates[1]
        self.h = DLH_rates[2]
        
        self.dupl_polytomy = dupl_polytomy
        
        self._reset()
        
        return self._gillespie_direct()
    
    
    def _reset(self):
        
        self.spec_queue = deque(self.sorted_speciations)    # queue for speciation events
        self.id_counter = 0
        
        self.total_rate = 0                                 # total event rate (all branches)
        self.branch_rates = []                              # array for branch rates
        
        self.ES_to_b = {e: [] for e in self.sorted_edges}   # maps E(S) --> existing branches
        self.i_to_b = {}                                    # maps array index --> branch
        self.b_to_i = {}                                    # maps branch --> array index

    
    def _get_tstamp(self, t):
        
        if self.total_rate <= 0.0:
            return -1
        else:
            return t - np.random.exponential(1/self.total_rate)
    
    
    def _get_branch_and_type(self):
        
        r = np.random.uniform() * self.total_rate
        index, current_sum = 0, 0
        
        for rate in self.branch_rates:
            if r <= current_sum + rate:
                break
            current_sum += rate
            index += 1
        
        branch = self.i_to_b[index]
        loss_factor = 1 if len(self.ES_to_b[(branch.S_u,branch.S_v)]) > 1 else 0
        
        if r <= current_sum + self.DLH_rates[0]:
            event_type = "D"
        elif r <= current_sum + self.DLH_rates[0] + loss_factor * self.DLH_rates[1]:
            event_type = "L"
        else:
            event_type = "H"
        
#        assert np.isclose(self.total_rate, sum(self.branch_rates)), "Sum of rates and total rate are not equal."
        
        return branch, event_type
    
    
    def _get_copy_number(self):
        
        if self.dupl_polytomy <= 0.0:
            return 2
        else:
            return 2 + np.random.poisson(lam=self.dupl_polytomy)


    def _coexisting_edges(self, tstamp, exclude_edge=None):
        """Return list of edges for the given timestamp."""
        
        valid_edges = []
        for edge in self.sorted_edges:
            if edge[0].tstamp <= tstamp:
                break
            elif edge[1].tstamp < tstamp and edge != exclude_edge:
                valid_edges.append(edge)
        return valid_edges


    def _gillespie_direct(self):
        
        T = self._initiatialize_tree()
        t = T.root.tstamp                   # start time = 1.0
        
        while self.spec_queue:
            
            event_tstamp = self._get_tstamp(t)
            next_spec_tstamp = self.spec_queue[0].tstamp
            
            # speciation
            if event_tstamp <= next_spec_tstamp:
                self._speciation()      
                t = next_spec_tstamp
                
            else:
                branch, event_type = self._get_branch_and_type()
                
                # duplication
                if event_type == "D":
                    self._duplication(event_tstamp, branch, event_type)
                
                # loss
                elif event_type == "L":
                    self._loss(event_tstamp, branch, event_type)
                    
                # HGT    
                elif event_type == "H":
                    self._hgt(event_tstamp, branch, event_type)
                
                t = event_tstamp

        return T
    
    
    def _initiatialize_tree(self):
        
        if len(self.S.root.children) > 1:
            # root is a speciation event
            T = PhyloTree(PhyloTreeNode(self.id_counter, label="S",
                                        color=self.S.root.ID, 
                                        dist=0.0,
                                        tstamp=self.S.root.tstamp))
        else:                    
            # planted species tree
            T = PhyloTree(PhyloTreeNode(self.id_counter,
                                        color=self.S.root.ID,
                                        dist=0.0,
                                        tstamp=self.S.root.tstamp))
        self.id_counter += 1
        self.spec_queue.popleft()
                
        for S_v in self.S.root.children:
            new_branch = Branch(self.id_counter, T.root, self.S.root, S_v, 0)
            self.ES_to_b[(self.S.root, S_v)].append(new_branch)
            index = len(self.branch_rates)
            self.branch_rates.append(self.d + self.h)
            self.total_rate += self.d + self.h
            self.i_to_b[index] = new_branch
            self.b_to_i[new_branch] = index
            self.id_counter += 1
            
        return T
            
    
    def _speciation(self):
        
        S_v = self.spec_queue.popleft()
        S_u = S_v.parent
        
        for branch in self.ES_to_b[(S_u, S_v)]:
            spec_node = PhyloTreeNode(branch.ID, label="S",
                                 color=S_v.ID, tstamp=S_v.tstamp,
                                 dist=abs(S_v.tstamp-branch.parent.tstamp),
                                 transferred=branch.transferred)
            branch.parent.add_child(spec_node)
            if not S_v.children:
                spec_node.label = str(spec_node.ID)
            for S_w in S_v.children:
                new_branch = Branch(self.id_counter, spec_node, S_v, S_w, 0)
                self.ES_to_b[(S_v, S_w)].append(new_branch)
                if S_w is S_v.children[0]:
                    index = self.b_to_i[branch]
                else:
                    index = len(self.branch_rates)
                    self.branch_rates.append(self.branch_rates[self.b_to_i[branch]])
                    self.total_rate += self.branch_rates[index]
                self.i_to_b[index] = new_branch
                self.b_to_i[new_branch] = index
                self.id_counter += 1
            del self.b_to_i[branch]
            
    
    def _duplication(self, event_tstamp, branch, event_type):
        
        S_u, S_v = branch.S_u, branch.S_v
        
        dupl_node = PhyloTreeNode(branch.ID, label="D",
                                  color=(S_u.ID,S_v.ID),
                                  tstamp=event_tstamp,
                                  dist=abs(event_tstamp-branch.parent.tstamp),
                                  transferred=branch.transferred)
        branch.parent.add_child(dupl_node)
        self.ES_to_b[(S_u, S_v)].remove(branch)
        
        copy_number = self._get_copy_number()
        
        for i in range(copy_number):
            new_branch = Branch(self.id_counter, dupl_node, S_u, S_v, 0)
            self.ES_to_b[(S_u, S_v)].append(new_branch)
            if i == 0:
                index = self.b_to_i[branch]
                self.branch_rates[index] = self.rate_sum
            else:
                index = len(self.branch_rates)
                self.branch_rates.append(self.rate_sum)
            self.i_to_b[index] = new_branch
            self.b_to_i[new_branch] = index
            self.id_counter += 1
        del self.b_to_i[branch]
        
        self.total_rate += (copy_number - 1) * (self.rate_sum)
        if len(self.ES_to_b[(S_u, S_v)]) == copy_number:
            self.total_rate += self.l
            
            
    def _loss(self, event_tstamp, branch, event_type):
        
        S_u, S_v = branch.S_u, branch.S_v
        
        loss_node = PhyloTreeNode(branch.ID, label="*",
                                  color=(S_u.ID,S_v.ID),
                                  tstamp=event_tstamp,
                                  dist=abs(event_tstamp-branch.parent.tstamp),
                                  transferred=branch.transferred)
        branch.parent.add_child(loss_node)
        self.ES_to_b[(S_u, S_v)].remove(branch)
        self.branch_rates[self.b_to_i[branch]] = 0
        del self.b_to_i[branch]
        
        if len(self.ES_to_b[(S_u, S_v)]) == 1:
            self.total_rate -= (self.rate_sum + self.l)
            self.branch_rates[self.b_to_i[self.ES_to_b[(S_u, S_v)][0]]] -= self.l
        else:
            self.total_rate -= self.rate_sum
    
    
    def _hgt(self, event_tstamp, branch, event_type):
        
        S_u, S_v = branch.S_u, branch.S_v
        
        valid_edges = self._coexisting_edges(event_tstamp,
                                             exclude_edge=(S_u, S_v))
        if valid_edges:
            trans_edge = random.choice(valid_edges)
            hgt_node = PhyloTreeNode(branch.ID, label="H",
                                     color=(S_u.ID,S_v.ID),
                                     tstamp=event_tstamp,
                                     dist=abs(event_tstamp-branch.parent.tstamp),
                                     transferred=branch.transferred)
            branch.parent.add_child(hgt_node)
            self.ES_to_b[(S_u, S_v)].remove(branch)
            
            # original branch
            new_branch = Branch(self.id_counter, hgt_node, S_u, S_v, 0)        
            self.ES_to_b[(S_u, S_v)].append(new_branch)
            index = self.b_to_i[branch]
            self.i_to_b[index] = new_branch
            self.b_to_i[new_branch] = index
            self.id_counter += 1
            
            # receiving branch
            trans_branch = Branch(self.id_counter, hgt_node, *trans_edge, 1)
            self.ES_to_b[trans_edge].append(trans_branch)
            index = len(self.branch_rates)
            self.branch_rates.append(self.rate_sum)
            self.i_to_b[index] = trans_branch
            self.b_to_i[trans_branch] = index
            self.id_counter += 1
            
            del self.b_to_i[branch]
            
            if len(self.ES_to_b[trans_edge]) == 2:
                self.total_rate += (self.rate_sum + self.l)
                self.branch_rates[self.b_to_i[self.ES_to_b[trans_edge][0]]] += self.l
            else:
                self.total_rate += self.rate_sum
                
    
    def _assert_no_extinction(self, T):
        """Returns False if gene family is extinct in some species."""
        
        VS_to_VT = {l.ID: [] for l in self.S.preorder() if not l.children}
        
        for v in T.preorder():
            if not v.children and not v.label == "*":
                VS_to_VT[v.color].append(v.ID)
                
        for leaf_list in VS_to_VT.values():
            if not leaf_list:
                return False
        
        return True
        

# --------------------------------------------------------------------------
#                 CONSTRUCTION OF THE OBSERVABLE TREE
#
# --------------------------------------------------------------------------
    
def observable_tree(tree):
    obs_tree = tree.copy()
    
    loss_nodes = []
    for node in obs_tree.postorder():
        if not node.children and node.label == "*":
            loss_nodes.append(node)
    
    # traverse from loss node to root delete if degree <= 1
    for loss_node in loss_nodes:
        current = obs_tree.delete_and_reconnect(loss_node,
                                                add_distances=True,
                                                keep_transferred=True)
        
        while len(current.children) < 2 and current.parent:
            current = obs_tree.delete_and_reconnect(current,
                                                    add_distances=True,
                                                    keep_transferred=True)
    
    # delete the root if the tree is planted
    if len(obs_tree.root.children) == 1:
        obs_tree.delete_and_reconnect(obs_tree.root.children[0],
                                      add_distances=False,
                                      keep_transferred=False)
    
    return obs_tree


# -*- coding: utf-8 -*-

import itertools

import networkx as nx

from asymmetree.cograph.Cograph import SimpleGraph
from asymmetree.cograph.CographEditor import CographEditor

from asymmetree.tools.PhyloTree import PhyloTree, PhyloTreeNode


__author__ = "David Schaller"
__copyright__ = "Copyright (C) 2020, David Schaller"


class TreeReconstructor:
    
    
    def __init__(self, edit_runs_per_cc=10, cotree_mode="best"):
        
        self.R = {}             # (weighted) species triples
        self.L = set()          # set of species leaves
        
        self._edit_runs_per_cc = edit_runs_per_cc   # runs of cograph editing heuristic
                                                    # per connected component
        self.S = None           # the reconstructed species tree
        
        if cotree_mode == "best":                   # only use the best cotree
            self._cotree_mode = "best"
        elif cotree_mode == "all":                  # use all cotrees
            self._cotree_mode = "all"
        else:
            raise ValueError("Invalid argument for cotree usage: '{}'.".format(cotree_mode))
        
        
    def add_ortho_graph(self, ortho_graph):
        
        for cc in nx.connected_components(ortho_graph):
            
            G = SimpleGraph()
            color_dict = {}
            
            for u in cc:
                G.add_node(u)
                for v in ortho_graph.neighbors(u):
                    G.add_edge(u, v)
                u_color = ortho_graph.nodes[u]['color']
                color_dict[u] = u_color
                self.L.add(u_color)
                
            # apply minimal cograph editing (heuristic)
            ce = CographEditor(G)
            cotree = ce.cograph_edit(run_number=self._edit_runs_per_cc)
            
            # add the informative triples (root is a speciation)
            if self._cotree_mode == "best":
                self._informative_triples(cotree, color_dict)
                
            elif self._cotree_mode == "all":
                weight_per_tree = 1 / len(ce.cotrees)
                for cotree in ce.cotrees:
                    self._informative_triples(cotree, color_dict,
                                              weight=weight_per_tree)
            
            
    def _informative_triples(self, cotree, color_dict, weight=1.0):
        """Add informative species triples from a (co)tree to R."""
        
        cotree.supply_leaves()
        
        for u in cotree.preorder():
            if u.label == "parallel" or not u.children:
                continue
            
            for v1, v2 in itertools.permutations(u.children, 2):
                for x, y in itertools.combinations(v1.leaves, 2):
                    for z in v2.leaves:
                        X, Y, Z = (color_dict[x.ID],
                                   color_dict[y.ID],
                                   color_dict[z.ID])
                        
                        if X != Y and X != Z and Y != Z:
                            self._add_triple(X, Y, Z, weight)
                            
        
    def _add_triple(self, a, b, c, weight):
        """Add a triple ab|c (= ba|c)."""
        
        if a <= b:
            triple = (a, b, c)
        else:
            triple = (b, a, c)
            
        if triple in self.R:
            self.R[triple] += weight
        else:
            self.R[triple] = weight
            
            
    def build_species_tree(self, mode='bpmf', weighted=True):
        """Build a species tree from the informative triples."""
        
        if mode.lower() == 'bpmf':
            root = self._BPMF(weighted=weighted)
        elif mode.lower() in ('mincut', 'min-cut'):
            root = self._BUILD(self.L, self.R, mincut=True)
        elif mode.lower() == 'greedy':
            root = self._GREEDY(weighted=weighted)
        else:
            raise ValueError("Mode '{}' is not valid!".format(mode))
            
        if not root:
            raise Exception("Could not build a species tree!")
        
        self.S = PhyloTree(root)
        
        return self.S
    
    
    def _BUILD(self, L, R, mincut=True):
        """Aho's BUILD-algorithm with minimal edge cut."""
        
        if len(L) == 1:                                 # trivial case: only one leaf left in L
            leaf = L.pop()
            return PhyloTreeNode(leaf, label=leaf)
        
        aho_graph = self._aho_graph(L, R)               # construct the Aho-graph
                                                        # determine connected components A1, ..., Ak
        conn_comps = self._connected_comp(aho_graph,
                                          mincut=mincut)
        
        if len(conn_comps) <= 1:                        # return False if less than 2 connected components
            return False
        
        child_nodes = []
        for cc in conn_comps:
            Li = set(cc)                                # list/dictionary --> set
            Ri = {}
            for t, weight in R.items():                 # determine which triples are in the subtree
                if Li.issuperset(t):
                    Ri[t] = weight
            Ti = self._BUILD(Li, Ri)                    # recursive call
            if not Ti:
                return False                            # raise False to previous call of aho()
            else:
                child_nodes.append(Ti)
                
        subtree_root = PhyloTreeNode(0, label="")       # place new inner node
        for Ti in child_nodes:
            subtree_root.add_child(Ti)                  # add roots of the subtrees to the new node
        
        return subtree_root                             # return the new node
    
    
    def _aho_graph(self, L, R):
        """Auxiliary graph for Aho-algorithm.
    
        Takes a list of leaves L and a list of tuples R of the form (xy|z) and
        builds a graph with V = L and E = { {x,y} | (xy|z) in R }.
        Applies Minimal Edge Cut if necessary.
        """
        G = nx.Graph()
        G.add_nodes_from(L)

        for a, b, c in R:
            if G.has_edge(a, b):
                G[a][b]["weight"] += R[(a,b,c)]
            else:
                G.add_edge(a, b, weight=R[(a,b,c)])
                
        return G
                
                
    def _connected_comp(self, G, mincut=True):
        
        if nx.number_connected_components(G) > 1 or not mincut:
            return list(nx.connected_components(G))
        else:
            cut = nx.stoer_wagner(G)                    # Stoer–Wagner algorithm
                                                        # for minimal weighted edge cut
            return cut[1]
        
    
    def _BPMF(self, weighted=True):
        """Wu’s Best-Pair-Merge-First heuristic.
        
        Modified version by Byrka et al. 2010 and added weights."""
        
        # initialization
        nodes = {PhyloTreeNode(leaf, label=leaf): {leaf} for leaf in self.L}
        leaf_to_node = {}
        
        for node in nodes:
            leaf_to_node[node.label] = node
        
        # merging
        for i in range(len(self.L)-1):
            
            score = {(S_i, S_j): 0
                     for S_i, S_j in itertools.combinations(nodes.keys(), 2)}
            
            for x, y, z in self.R.keys():
                
                w = self.R[(x,y,z)]
                
                S_i, S_j, S_k = (leaf_to_node[x],
                                 leaf_to_node[y],
                                 leaf_to_node[z])
                
                if (S_i is not S_j) and (S_i is not S_k) and (S_j is not S_k):
                    
                    if (S_i, S_j) in score:
                        score[(S_i, S_j)] += 2 * w if weighted else 2
                    else:
                        score[(S_j, S_i)] += 2 * w if weighted else 2
                        
                    if (S_i, S_k) in score:
                        score[(S_i, S_k)] -= 1 * w if weighted else 1
                    else:
                        score[(S_k, S_i)] -= 1 * w if weighted else 1
                        
                    if (S_j, S_k) in score:
                        score[(S_j, S_k)] -= 1 * w if weighted else 1
                    else:
                        score[(S_k, S_j)] -= 1 * w if weighted else 1
            
            current_max = float('-inf')
            S_i, S_j = None, None
            
            for pair, pair_score in score.items():
                
                if pair_score > current_max:
                    current_max = pair_score
                    S_i, S_j = pair
            
            # create new node S_k connecting S_i and S_j
            S_k = PhyloTreeNode(0, label="")
            S_k.add_child(S_i)
            S_k.add_child(S_j)
            
            nodes[S_k] = nodes[S_i] | nodes[S_j]    # set union
            for leaf in nodes[S_k]:
                leaf_to_node[leaf] = S_k
            
            del nodes[S_i]
            del nodes[S_j]
            
        assert len(nodes) == 1, "More than 1 node left!"
        
        return next(iter(nodes))
    
    
    def _GREEDY(self, weighted=True):
        
        if weighted:
            triples = sorted(self.R.keys(),
                             key=lambda triple: self.R[triple],
                             reverse=True)
        else:
            triples = self.R.keys()
                
        consistent_triples = {}
        root = self._BUILD(self.L, consistent_triples, mincut=False)
        
        for t in triples:
            consistent_triples[t] = self.R[t]
            
            build_tree = self._BUILD(self.L, consistent_triples, mincut=False)
            if build_tree:
                root = build_tree
            else:
                del consistent_triples[t]
        
        return root
    
    
    def _max_consistent_triple_set(self):
        
        if self.S is None:
            raise Exception("Species tree has not been built yet!")
            
        R_total = self.S.get_triples()      # all triples of the tree
        R_max_cons = {}                     # max. consistent subset of R (i.e. heuristic)
        
        for l1, l2, l3 in R_total:
            a = l1.label
            b = l2.label
            c = l3.label
            
            if a <= b:
                triple = (a, b, c)
            else:
                triple = (b, a, c)
                
            if triple in self.R:
                R_max_cons[triple] = self.R[triple]
        
        return R_max_cons
    
    
    def _total_support(self, R_max_cons):
        """Compute the total support s of the tree as in ParaPhylo."""
        
        numerator = 0
        denominator = 0
        
        for triple, weight in R_max_cons.items():
            
            numerator += weight
            denominator += weight
            
            a, b, c = triple
            triple2 = (a, c, b) if a <= c else (c, a, b)
            triple3 = (b, c, a) if b <= c else (c, b, a)
            
            if triple2 in self.R:
                denominator += self.R[triple2]
            if triple3 in self.R:
                denominator += self.R[triple3]
        
        support = numerator / denominator if denominator > 0 else 0
        return support
    
    
    def _subtree_support(self):
        """Compute the subtree support s_v for the inner nodes as in ParaPhylo."""
        
        support = {}                        # internal node v --> support for T(v)
        
        self.S.supply_leaves()
        all_leaves = set(self.S.root.leaves)
        
        for v in self.S.preorder():
            if not v.children or v is self.S.root:
                continue
            
            numerator = 0
            denominator = 0
            outspecies = all_leaves.difference(v.leaves)
            
            for l1, l2 in itertools.combinations(v.leaves, 2):
                for l3 in outspecies:
                    a = l1.label
                    b = l2.label
                    c = l3.label
                    triple1 = (a, b, c) if a <= b else (b, a, c)
                    triple2 = (a, c, b) if a <= c else (c, a, b)
                    triple3 = (b, c, a) if b <= c else (c, b, a)
                    
                    if triple1 in self.R:
                        numerator += self.R[triple1]
                        denominator += self.R[triple1]
                    if triple2 in self.R:
                        denominator += self.R[triple2]
                    if triple3 in self.R:
                        denominator += self.R[triple3]
                        
            support[v] = numerator / denominator if denominator > 0 else 0
        
        return support
    
    
    def newick_with_support(self, v=None, supports=None):
        """Recursive PhyloTree --> Newick (str) function."""
        
        if v is None:
            supports = self._subtree_support()
            supports[self.S.root] = self._total_support(self._max_consistent_triple_set())
            return self.newick_with_support(v=self.S.root, supports=supports) + ";"
        
        elif not v.children:
            return str(v.label)
        
        else:
            s = ''
            for child in v.children:
                s += self.newick_with_support(v=child, supports=supports) + ","
            return "({}){}".format(s[:-1], supports[v])
        

if __name__ == "__main__":
    
    # ----- TESTING THIS MODULE -----
    
    import asymmetree.simulator.TreeSimulator as ts
    import asymmetree.simulator.TreeImbalancer as tm
    from asymmetree.best_matches import TrueBMG
    
    # SPECIES TREE:
    #S = ts.build_species_tree(10)
    S = PhyloTree.parse_newick("(((((16:0.38786287055727103,(18:0.2071277923445058,19:0.2071277923445058)17:0.18073507821276524)12:0.10075553853805931,(14:0.13224182895383052,15:0.13224182895383052)13:0.3563765801414998)4:0.07517286794939665,(6:0.5373882998574596,(8:0.4434182448023457,(10:0.04929450217312242,11:0.04929450217312242)9:0.3941237426292233)7:0.0939700550551139)5:0.02640297718726732)2:0.2512472266526016,3:0.8150385036973286)1:0.18496149630267142)0:0.0;")
    S.reconstruct_IDs()
    S.reconstruct_timestamps()
    
    print(S.to_newick())
    
    tr = TreeReconstructor(cotree_mode="best")
    
    # GENE FAMILIES:
    for i in range(100):
        TGT_simulator = ts.GeneTreeSimulator(S)
        TGT = TGT_simulator.simulate((1,1,0))
        TGT = tm.imbalance_tree(TGT, S, baseline_rate=1,
                                lognormal_v=0.2,
                                gamma_param=(0.5, 1.0, 2.2),
                                weights=(1, 1, 1))
        OGT = ts.observable_tree(TGT)
        
    #    ortho_graph = TrueBMG.true_orthology_graph(OGT)
    #    tr.add_ortho_graph(ortho_graph)
        
        _, RBMG = TrueBMG.best_match_graphs(OGT)
        tr.add_ortho_graph(RBMG)
        
        
    S_estimate = tr.build_species_tree(mode="mincut")
    print(tr.newick_with_support())
    
    S_estimate2 = tr.build_species_tree(mode="BPMF")
    print(tr.newick_with_support())


# -*- coding: utf-8 -*-

from Bio import pairwise2
from Bio.SubsMat import MatrixInfo

from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.Align import MultipleSeqAlignment

from Bio.Phylo.TreeConstruction import DistanceCalculator


def distance_2seqs(seq1, seq2, matrix=MatrixInfo.blosum62,
                   model='blosum62'):

    aln = pairwise2.align.globaldx(seq1, seq2, matrix,
                                   one_alignment_only=True)
    
    # DistanceCalculator only accepts MultipleSeqAlignment
    msa = MultipleSeqAlignment([SeqRecord(Seq(aln[0][0]), id="seq1"),
                                SeqRecord(Seq(aln[0][1]), id="seq2"),
                                ])
    
    calculator = DistanceCalculator(model)
    dm = calculator.get_distance(msa)
    
    return dm.matrix[1][0]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="asymmetree",
    version="0.0.5",
    author="David Schaller",
    author_email="david@0x002A.com",
    description="Simulation of species and gene tree scenarios with asymmetric evolution rates.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/david-schaller/asymmetree",
    packages=setuptools.find_packages(exclude=["tests", "tests.*"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'numpy>=1.16.4',
        'scipy>=1.3.0',
        'matplotlib>=3.0',
        'networkx>=2.2',
   ],
)



# -*- coding: utf-8 -*-

"""
Tree data structure for phylogentic trees.

Classes in this module:
    - PhyloTreeNode
    - PhyloTree
"""

import collections, random, re
import networkx as nx

from asymmetree.tools.Tree import Tree, TreeNode

__author__ = "David Schaller"
__copyright__ = "Copyright (C) 2020, David Schaller"


class PhyloTreeNode(TreeNode):
    """Class 'PhyloTreeNode'.
    
    Components for class 'PhyloTree'.
    """
    
    __slots__ = ('color', 'dist', 'tstamp', 'transferred')
    
    
    def __init__(self, ID, label="", color=None,
                 dist = 1.0, tstamp=None, transferred=0):
        
        super().__init__(ID, label=label)
        self.color = color
        self.dist = dist
        self.tstamp = tstamp
        self.transferred = transferred

    
    def __repr__(self):
        
        return "pn" + str(self.ID)
    
    
    def __str__(self):
        
        if isinstance(self.color, tuple):
            return "{}<{}-{}>:{}".format(self.label, *self.color, self.dist)
        elif self.color:
            return "{}<{}>:{}".format(self.label, self.color, self.dist)
        else:
            return "{}:{}".format(self.label, self.dist)


class PhyloTree(Tree):
    
    # corresponding node type
    node_type = PhyloTreeNode
    
    
    def __init__(self, root):
        
        if not isinstance(root, PhyloTreeNode):
            raise TypeError("root must be of type 'PhyloTreeNode'.")
        super().__init__(root)
    
    
    def sorted_nodes(self):
        """Return list of sorted nodes by timestamp."""
        
        return sorted(self.preorder(),
                      key=lambda node: node.tstamp,
                      reverse=True)
    
    
    def sorted_edges(self):
        """Return list of sorted edges (u,v) by timestamp of u."""
    
        return sorted(self.edges(),
                      key=lambda e: (e[0].tstamp, e[1].tstamp),
                      reverse=True)
    
    
    def delete_and_reconnect(self, node,
                             add_distances=True,
                             keep_transferred=True):
        """Delete a node from the tree and reconnect its parent and children."""
        
        distance, transferred = node.dist, node.transferred
        parent = super().delete_and_reconnect(node)
        
        if parent:
            for child in parent.children:
                if add_distances:
                    child.dist += distance
                if keep_transferred and transferred:
                    child.transferred = 1
        
        return parent
        
    
    def contract(self, edges):
        
        contracted = set()
        
        for u, v in edges:
            
            # avoid trying to contract the same edge multiple times
            if v not in contracted:
                self.delete_and_reconnect(v)
            
            contracted.add(v)
            
    
    def supply_leaves(self, node=None, exclude_losses=True):
        """Add the leaves to all nodes that are in the subtree of a specific node."""
        
        if not node:
            self.supply_leaves(self.root, exclude_losses=exclude_losses)
            return self.root.leaves
        else:
            node.leaves = []
            if not node.children:
                if exclude_losses and node.label != "*":
                    node.leaves.append(node)
                elif not exclude_losses:
                    node.leaves.append(node)
            else:
                for child in node.children:
                    node.leaves.extend(self.supply_leaves(child, exclude_losses=exclude_losses))
            return node.leaves
    
    
    def distances_from_root(self):
        
        leaf_distances = []
        distance_dict = {}
        for v in self.preorder():
            if not v.parent:
                distance_dict[v] = 0.0
            else:
                depth = distance_dict[v.parent] + v.dist
                distance_dict[v] = depth
            if not v.children:
                leaf_distances.append( (str(v), distance_dict[v]) )
        return distance_dict, leaf_distances
    
    
    def topology_only(self, inplace=True):
        """Reset distances, time stamps, transfer status and inner labels."""
        
        if not inplace:
            T = self.copy()
        else:
            T = self
        
        for v in T.preorder():
            if v.children:
                v.label = ""
                v.color = None
            v.dist = 1.0
            v.tstamp = None
            v.transferred = 0
        
        return T
    

# --------------------------------------------------------------------------
#                          TREE  <--->  NEWICK
# --------------------------------------------------------------------------
        
    def to_newick(self, node=None, 
                  label=True, color=True, distance=True,
                  label_inner=True, color_inner=False):
        """Recursive PhyloTree --> Newick (str) function."""
        
        if node is None:
            return self.to_newick(self.root,
                                  label=label, color=color, distance=distance,
                                  label_inner=label_inner, color_inner=color_inner) + ";"
        elif not node.children:
            token = ""
            if label:
                token += str(node.label)
            if color and node.color:
                token += "<{}-{}>".format(*node.color) if isinstance(node.color, tuple) else "<{}>".format(node.color)
            if distance:
                token += ":{}".format(node.dist)
            return token
        else:
            s = ''
            for child in node.children:
                s += self.to_newick(node=child,
                                    label=label, color=color, distance=distance,
                                    label_inner=label_inner, color_inner=color_inner) + ","
            token = ""
            if label and label_inner:
                token += str(node.label)
            if color_inner and node.color:
                token += "<{}-{}>".format(*node.color) if isinstance(node.color, tuple) else "<{}>".format(node.color)
            if distance:
                token += ":{}".format(node.dist)
            return "({}){}".format(s[:-1], token)
    
    
    @staticmethod
    def parse_newick(newick):
        """Parses trees in Newick format into object of type 'PhyloTree'."""
        
        # label<color>:distance
        label_col_dist_regex = re.compile(r"'?([a-zA-Z0-9_]*)'?<(.*)>:(-?[0-9]*\.?[0-9]*[Ee]?-?[0-9]+)")
        # label<color>
        label_col_regex = re.compile(r"'?([a-zA-Z0-9_]*)'?<(.*)>")
        # label:distance
        label_dist_regex = re.compile(r"'?([a-zA-Z0-9_]*)'?:(-?[0-9]*\.?[0-9]*[Ee]?-?[0-9]+)")
        
        id_counter = 0
        
        def parse_subtree(subroot, subtree_string):
            """Recursive function to parse the subtrees."""
            nonlocal id_counter
            children = split_children(subtree_string)
            for child in children:
                node = PhyloTreeNode(id_counter)
                subroot.add_child(node)
                id_counter += 1
                end = -1
                if child[0] == '(':                                 # the child has subtrees
                    end = child.rfind(')')
                    if end == -1:
                        raise ValueError("Invalid Newick-String!")
                    parse_subtree(node, child[1:end])               # recursive call 'parse_subtree'
                child = child[end+1:].strip()
                label_col_dist = label_col_dist_regex.match(child)
                if label_col_dist:                                  # CASE 1: label<color>:distance
                    node.label = label_col_dist.group(1)
                    node.color = label_col_dist.group(2)
                    node.dist = float(label_col_dist.group(3))
                else:
                    label_col = label_col_regex.match(child)
                    label_dist = label_dist_regex.match(child)
                    if label_col:                                   # CASE 2: label<color>
                        node.label = label_col.group(1)
                        node.color = label_col.group(2)
                    elif label_dist:                                # CASE 3: label:distance
                        node.label = label_dist.group(1)
                        node.dist = float(label_dist.group(2))
                    else:                                           # CASE 4: label
                        node.label = child
                if node.color and node.color.find('-') != -1:
                    split_color = node.color.split('-')
                    node.color = (split_color[0], split_color[1])
                        
        def split_children(child_string):
            """Splits a given string by all ',' that are not enclosed by parentheses."""
            
            stack = 0
            children = []
            current = ""
            for c in child_string:
                if (stack == 0) and c == ',':
                    children.append(current)
                    current = ""
                elif c == '(':
                    stack += 1
                    current += c
                elif c == ')':
                    if stack <= 0:
                        raise ValueError("Invalid Newick-String!")
                    stack -= 1
                    current += c
                else:
                    current += c
            children.append(current.strip())
            return children
        
        if not isinstance(newick, str):
            raise ValueError("Newick parser needs a 'str' as input.")
        end = newick.find(";")
        if end != -1:
            newick = newick[:end]
        temp_root = PhyloTreeNode(-1)
        parse_subtree(temp_root, newick)
        if temp_root.children:
            root = temp_root.children[0]
            root.dist = 0.0                 # set distance of the root to 0
            root.detach()                   # remove the parent temp_root
                                            # (important for non-recursive to_newick2)
            return PhyloTree(root)
        else:
            raise ValueError("Invalid Newick-String!")
    
    
    def reconstruct_IDs(self):
        """Reconstruct the (leaf) IDs."""
        
        self.number_of_species = 0
        IDs = set()
        
        for v in self.preorder():
            if not v.children:
                self.number_of_species += 1
            if v.label.isdigit():
                v.ID = int(v.label)
                IDs.add(v.ID)
        
        # assign new IDs to internal nodes
        current_ID = 0
        for v in self.preorder():
            if str(v.ID) != v.label:
                while current_ID in IDs:
                    current_ID += 1
                v.ID = current_ID
                IDs.add(current_ID)
                
                
    def reconstruct_timestamps(self):
        """Reconstruct the timestamps."""
        
        self.root.tstamp = 1.0
        for v in self.preorder():
            if v.parent:
                v.tstamp = v.parent.tstamp - v.dist
    
# --------------------------------------------------------------------------
#                         TREE  <--->  NETWORKX
# --------------------------------------------------------------------------
            
    def to_nx(self):
        
        self.check_integrity()
        G = nx.DiGraph()
        
        for u, v in self.edges():
            if u.ID not in G:
                G.add_node(u.ID, label=u.label, color=u.color, tstamp=u.tstamp)
            if v.ID not in G:
                G.add_node(v.ID, label=v.label, color=v.color, tstamp=v.tstamp)
            if u is v or u.ID == v.ID:
                print("Loop at", str(u), str(v), v.transferred)
            G.add_edge(u.ID, v.ID, dist=v.dist, transferred=v.transferred)
            
        return G, self.root.ID
    
    
    @staticmethod
    def parse_nx(G, root):
        
        number_of_leaves = 0
    
        def build_tree(ID, parent=None):
            nonlocal number_of_leaves
            treenode = PhyloTreeNode(ID, label=G.nodes[ID]['label'])
            if parent:
                parent.add_child(treenode)
                if 'dist' in G.edges[parent.ID, ID]:
                    treenode.dist = G.edges[parent.ID, ID]['dist']
                if 'transferred' in G.edges[parent.ID, ID]:
                    treenode.transferred = G.edges[parent.ID, ID]['transferred']
            
            if 'label' in G.nodes[ID]:
                treenode.label = G.nodes[ID]['label']
            if 'color' in G.nodes[ID]:
                treenode.color = G.nodes[ID]['color']
            if 'tstamp' in G.nodes[ID]:
                treenode.tstamp = G.nodes[ID]['tstamp']
            
            for child_id in G.neighbors(ID):
                build_tree(child_id, parent=treenode)
            if G.out_degree(ID) == 0:
                number_of_leaves += 1
            
            return treenode
        tree = PhyloTree(build_tree(root))
        tree.number_of_species = number_of_leaves
        
        return tree
# --------------------------------------------------------------------------

    
    def copy(self):
        
        orig_to_new = {}
        
        for orig in self.preorder():
            new = PhyloTreeNode(orig.ID, label=orig.label, 
                                color=orig.color, dist=orig.dist,
                                tstamp=orig.tstamp,
                                transferred=orig.transferred)
            orig_to_new[orig] = new
            if orig.parent:
                orig_to_new[orig.parent].add_child(new)
        
        return PhyloTree(orig_to_new[self.root])
    
    
    @staticmethod
    def random_colored_tree(N, colors, binary=False):
        """Creates a random colored tree.
        
        The number of leaves and the color labels are specified in the
        parameters 'N' and 'colors', respectively. Each non-leaf node in the 
        resulting tree will have at least children (property of phylogenetic
        trees).
        
        Keyword arguments:
            binary - forces the tree to be binary; default is False
        """
        
        if not (isinstance(N, int) and isinstance(colors, collections.Iterable)):
            raise TypeError("N must be of type 'int' and colors must be iterable!")
        root = PhyloTreeNode(0, label="0")
        tree = PhyloTree(root)
        node_list = [root]
        break_prob = [0.5, 0.5]
        nr, leaf_count = 1, 1
        
        while leaf_count < N:
            node = random.choice(node_list)
            
            if not node.children: 
                # to be phylogenetic at least two children must be added
                new_child1 = PhyloTreeNode(nr, label=str(nr))
                new_child2 = PhyloTreeNode(nr+1, label=str(nr+1))
                node.add_child(new_child1)
                node.add_child(new_child2)
                node_list.extend(node.children)
                nr += 2
                leaf_count += 1
            elif node.children and not binary:
                # add only one child if there are already children
                
                # probability to add another child (decraeses exponentially)
                break_prob[0] = 1 / 1.4**len(node.children)
                # probability to choose another node
                break_prob[1] = 1 - break_prob[0]
                if random.choices( (0,1), weights=break_prob )[0] == 1:
                    continue
                
                new_child = PhyloTreeNode(nr, label=str(nr))
                node.add_child(new_child)
                node_list.append(new_child)
                nr += 1
                leaf_count += 1
                
        for node in node_list:                          # assign colors
            if not node.children:                       # to leaves randomly
                node.color = random.choice(colors)
                
        return tree
    
    
    def check_integrity(self):
        
        for v in self.preorder():
            for child in v.children:
                if child is v:
                    print("Loop at " + str(v))
                    raise KeyboardInterrupt
                if child.parent is not v:
                    print("PhyloTree invalid for " + str(v) + " and " + str(child))
                    raise KeyboardInterrupt

    
if __name__ == "__main__":
    colors = ("s", "t", "v", "w")
    N = 20
    
    t = PhyloTree.random_colored_tree(N, colors)
    print("------------- Random tree test -------------")
    print( t.to_newick() )
    print("--------------------------------------------")
    
    t2 = PhyloTree.parse_newick(t.to_newick())
    print( t2.to_newick() )
    
    nx_tree, nx_root = t.to_nx()
    t3 = PhyloTree.parse_nx(nx_tree, nx_root)
    print("--------------------------------------------")
    print( t3.to_newick() )



# -*- coding: utf-8 -*-

"""
Linear cograph detector.

Implementation of:
    D. G. Corneil, Y. Perl, and L. K. Stewart.
    A Linear Recognition Algorithm for Cographs.
    SIAM J. Comput., 14(4), 926–934 (1985).
    DOI: 10.1137/0214065
"""

from collections import deque

from asymmetree.cograph.Cograph import Cotree, CotreeNode


__author__ = "David Schaller"
__copyright__ = "Copyright (C) 2019, David Schaller"


class LCD:
    
    def __init__(self, G):
        
        self.G = G
        self.V = list(G.adj_list.keys())
        
        self.T = Cotree(None)
        self.already_in_T = set()
        self.leaf_map = {}
        self.node_counter = 0
        
        self.marked = set()
        self.m_u_children = {}              # lists of marked and unmarked children
        self.mark_counter = 0
        self.unmark_counter = 0
        
        self.error_message = ""
    
    
    def cograph_recognition(self):
        
        if len(self.V) == 0:
            print("Empty graph in Cograph Recognition!")
            return self.T
        
        elif len(self.V) == 1:
            self.T.root = CotreeNode(self.V[0], label="leaf")
            return self.T
        
        v1, v2 = self.V[0], self.V[1]
        self.already_in_T.update([v1, v2])
        
        R = CotreeNode(None, label="series")
        self.T.root = R
        
        if self.G.has_edge(v1, v2):
            v1_node = CotreeNode(v1, label="leaf")
            v2_node = CotreeNode(v2, label="leaf")
            R.add_child(v1_node)
            R.add_child(v2_node)
            self.node_counter = 3
        else:
            N = CotreeNode(None, label="parallel")
            R.add_child(N)
            v1_node = CotreeNode(v1, label="leaf")
            v2_node = CotreeNode(v2, label="leaf")
            N.add_child(v1_node)
            N.add_child(v2_node)
            self.node_counter = 4
            
        self.leaf_map[v1] = v1_node
        self.leaf_map[v2] = v2_node
        
        if len(self.V) == 2:
            return self.T
        
        for x in self.V[2:]:
            
#            print(str(x)+" before "+self.T.to_newick())
            
            # initialization (necessary?)
            self.marked.clear()
            self.m_u_children.clear()
            self.mark_counter = 0
            self.unmark_counter = 0
            self.already_in_T.add(x)        # add x for subsequent iterations
            
            # call procedure _MARK(x)
            self._MARK(x)
            
            # all nodes in T were marked and unmarked
            if self.node_counter == self.unmark_counter:
                R = self.T.root
                x_node = CotreeNode(x, label="leaf")
                R.add_child(x_node)
                self.node_counter += 1
                self.leaf_map[x] = x_node
                continue
            # no nodes in T were marked and unmarked
            elif self.mark_counter == 0:
                # d(R)=1
                if len(self.T.root.children) == 1:
                    N = self.T.root.children[0]
                    x_node = CotreeNode(x, label="leaf")
                    N.add_child(x_node)
                    self.node_counter += 1
                else:
                    R_old = self.T.root
                    R_new = CotreeNode(None, label="series")
                    N = CotreeNode(None, label="parallel")
                    R_new.add_child(N)
                    N.add_child(R_old)
                    self.T.root = R_new
                    
                    x_node = CotreeNode(x, label="leaf")
                    N.add_child(x_node)
                    self.node_counter += 3
                self.leaf_map[x] = x_node
                continue
            
            u = self._find_lowest()
            if not u:
                return False
#            print(u)
            
            # label(u)=0 and |A|=1
            if u.label == "parallel" and len(self.m_u_children[u]) == 1:
                w = self.m_u_children[u][0]
                if w.label == "leaf":
                    new_node = CotreeNode(None, label="series")
                    u.remove_child(w)
                    u.add_child(new_node)
                    new_node.add_child(w)
                    
                    x_node = CotreeNode(x, label="leaf")
                    new_node.add_child(x_node)
                    self.node_counter += 2
                else:
                    x_node = CotreeNode(x, label="leaf")
                    w.add_child(x_node)
                    self.node_counter += 1 
            
            # label(u)=1 and |B|=1
            elif (u.label == "series" and 
                  len(u.children) - len(self.m_u_children[u]) == 1):
                set_A = set(self.m_u_children[u])       # auxiliary set bounded by O(deg(x))
                w = None
                for child in u.children:
                    if child not in set_A:
                        w = child
                        break
                if w.label == "leaf":
                    new_node = CotreeNode(None, label="parallel")
                    u.remove_child(w)
                    u.add_child(new_node)
                    new_node.add_child(w)
                    
                    x_node = CotreeNode(x, label="leaf")
                    new_node.add_child(x_node)
                    self.node_counter += 2
                else:
                    x_node = CotreeNode(x, label="leaf")
                    w.add_child(x_node)
                    self.node_counter += 1
            
            else:
                y = CotreeNode(None, label=u.label)
                for a in self.m_u_children[u]:
                    u.remove_child(a)
                    y.add_child(a)
                    
                if u.label == "parallel":
                    new_node = CotreeNode(None, label="series")
                    u.add_child(new_node)
                    
                    new_node.add_child(y)
                    x_node = CotreeNode(x, label="leaf")
                    new_node.add_child(x_node)
                else:
                    par = u.parent
                    if par is not None:             # u was the root of T
                        par.remove_child(u)
                        par.add_child(y)
                    else:
                        self.T.root = y             # y becomes the new root
                    
                    new_node = CotreeNode(None, label="parallel")
                    y.add_child(new_node)
                    new_node.add_child(u)
                    x_node = CotreeNode(x, label="leaf")
                    new_node.add_child(x_node)
                self.node_counter += 3
                
            self.leaf_map[x] = x_node
        
        return self.T
    
    
    def _MARK(self, x):
        
        for v in self.G.neighbors(x):
            if v in self.already_in_T:
                self.marked.add(self.leaf_map[v])
                self.mark_counter += 1
                
        queue = deque(self.marked)
        
        while queue:                        # contains only d(u)=md(u) nodes
            u = queue.popleft()
            self.marked.remove(u)           # unmark u
            self.unmark_counter += 1
            u.aux_counter = 0               # md(u) <- 0
            if u is not self.T.root:
                w = u.parent                # w <- parent(u)
                if w not in self.marked:
                    self.marked.add(w)      # mark w
                    self.mark_counter += 1
                w.aux_counter += 1
                if w.aux_counter == len(w.children):
                    queue.append(w)
                    
                if w in self.m_u_children:              # append u to list of
                    self.m_u_children[w].appendleft(u)  # marked and unmarked
                else:                                   # children of w
                    self.m_u_children[w] = deque([u])
                    
        if (self.marked and                             # any vertex remained marked
            len(self.T.root.children) == 1 and 
            self.T.root not in self.marked):
            
            self.marked.add(self.T.root)
            self.mark_counter += 1
    
    
    def _find_lowest(self):
        
        R = self.T.root
        y = "Lambda"
        
        if R not in self.marked:        # R is not marked
            self.error_message = "(iii): R=" + str(R)
            return False                # G+x is not a cograph (iii)
        else:
            if R.aux_counter != len(R.children) - 1:
                y = R
            self.marked.remove(R)
            R.aux_counter = 0
            u = w = R
        
        while self.marked:              # while there are mark vertices
            u = self.marked.pop()       # choose a arbitrary marked vertex u
            
            if y != "Lambda":
                self.error_message = "(i) or (ii): y=" + str(y)
                return False            # G+x is not a cograph (i) or (ii)
            
            if u.label == "series":
                if u.aux_counter != len(u.children) - 1:
                    y = u
                if u.parent in self.marked:
                    self.error_message = "(i) and (vi): u=" + str(u)
                    return False        # G+x is not a cograph (i) and (vi)
                else:
                    t = u.parent.parent
            else:
                y = u
                t = u.parent
            u.aux_counter = 0           # u was already unmarked above
            
            # check if the u-w path is part of the legitimate alternating path
            while t is not w:
                if t is R:
                    self.error_message = "(iv): t=" + str(t)
                    return False        # G+x is not a cograph (iv)
                
                if t not in self.marked:
                    self.error_message = "(iii), (v) or (vi): t=" + str(t)
                    return False        # G+x is not a cograph (iii), (v) or (vi)
                
                if t.aux_counter != len(t.children) - 1:
                    self.error_message = "(ii): t=" + str(t)
                    return False        # G+x is not a cograph (ii)
                
                if t.parent in self.marked:
                    self.error_message = "(i): t=" + str(t)
                    return False        # G+x is not a cograph (i)
                
                self.marked.remove(t)   # unmark t
                t.aux_counter = 0       # reset md(t)
                t = t.parent.parent
                
            w = u                       # rest w for next choice of marked vertex
        
        return u


if __name__ == "__main__":
    
    from asymmetree.cograph.Cograph import SimpleGraph
    cotree = Cotree.random_cotree(100)
#    print(cotree.to_newick())
    cograph = cotree.to_cograph()
    
#    # (1,((5,(7,8)<0>)<1>,4)<0>)<1>;
#    cograph = SimpleGraph(initial={1: {8, 4, 5, 7}, 5: {8, 1, 7}, 7: {1, 5}, 8: {1, 5}, 4: {1}})
    #print(cograph.adj_list)
    
    LCD = LCD(cograph)
    new_cotree = LCD.cograph_recognition()
    print("done")
    if new_cotree:
#        print(new_cotree.to_newick())
        new_cograph = new_cotree.to_cograph()
        #print(new_cograph.adj_list)
        print(cograph.graphs_equal(new_cograph))
    else:
        print("Not a cograph!")
    